using System;
using System.Drawing;		// for Bitmap and Color
using System.Collections;   // for ArrayList
using System.Threading;     // for Sleep()
using System.Runtime.Serialization;

namespace epitome
{
	/// <summary>
	/// Summary description for Epitome.
	/// </summary>
	[Serializable]
	public class Epitome : IDeserializationCallback
	{
		public Bitmap EpitomeBitmap;
		public Bitmap LatestBitmap; // a readable copy of EpitomeBitmap
		public bool LatestBitmapOk; // flag for whether it's ok to ask for the latest bitmap.
		[NonSerialized] public Bitmap PatchQuilt;	// a readable bitmap containing a sample of the patches.
		[NonSerialized] public bool PatchesOk;		// a flag for whether it's ok to ask for the quilt yet.

		public ArrayList Patches;
		public int NumPatches;

		// The following two maps support faster EpitomeBitmap updating, epitome-to-source remapping, plus
		// interactive features for finding out exactly where the image-to-epitome correspondences are.  The 
		// interactive features are a lot easier if these ArrayLists are public.

		// EpitomePatchMap is a map from locations in the epitome to an ArrayList of patches at that location.
		// Every time we update the DestinationX,Y fields in a patch, we have to update EpitomePatchMap as well.
		public ArrayList[,] EpitomePatchMap;
		// SourcePatchMap is a map from locations in the source to an ArrayList of patches at that location.
		// Every time we update the OriginX,Y fields in a patch, we have to update the SourcePatchMap as well.
		public ArrayList[,] SourcePatchMap;
		private static int MapAdd = 0;     // numeric aliases for the actions of adding and removing from a 
		private static int MapRemove = 1;  // patch map.

		[NonSerialized] public ArrayList EpitomeSequence; // an array of epitome bitmaps, one for each iteration.
		private static int MaxSequenceBitmaps = 40;

		public string StatString;

		[NonSerialized] private Random R;
		[NonSerialized] private bool StopRequested;

		// fields to capture input parameters from the UI.
		private Bitmap Source;
		private int Width, Height,
			PatchSizeMin, PatchSizeMax,
			MaxIterations, RandomIterationModulus,
			QuiltWidth, QuiltHeight;
		public int W { get { return Width; } }
		public int H { get { return Height; } }
		public int[] PatchSizes { get { return new int[] {PatchSizeMin, PatchSizeMax}; } }
		private string PixelErrorMethod;
		public string ErrorMethod 
		{
			get { return PixelErrorMethod; }
		}

		// two EventArgs structures we can use to pass strings and bitmaps back to the UI.
		public class EpitomeStringArgs : EventArgs 
		{
			public string status;
			public EpitomeStringArgs(string s) 
			{
				status = s;
			}
		}
		public class EpitomeBitmapArgs : EventArgs 
		{
			public Bitmap bitmap;
			public EpitomeBitmapArgs(Bitmap b) 
			{
				bitmap = b;
			}
		}

		// two delegates corresponding to the two EventArgs structures:
		public delegate void EpitomeStringEvent(object o, EpitomeStringArgs e);
		public delegate void EpitomeBitmapEvent(object o, EpitomeBitmapArgs e);
		// and a third parameter-less event to signal the UI thread when we're done.
		public delegate void EpitomeDoneEvent(object o, EventArgs e);

		// events corresponding to those delegates:
		[field: NonSerialized] public event EpitomeStringEvent OnEpitomeString;
		[field: NonSerialized] public event EpitomeBitmapEvent OnEpitomeBitmap;
		[field: NonSerialized] public event EpitomeDoneEvent   OnEpitomeDone;

		public Epitome(Bitmap source, int width, int height,
			int patchsizemax, bool fixedsize,
			int maxiterations, int randomiterationmodulus,
			string errormethod,
			int quiltwidth, int quiltheight) 
		{
			Source = source;				// remember all these parameters for use by GenerateEpitome()
			Width = width;
			Height = height;

			// initialize the patch maps.
			EpitomePatchMap = new ArrayList[Width, Height];
			for(int i = 0; i < Width; i++)
				for(int j = 0; j < Height; j++)
					EpitomePatchMap[i,j] = new ArrayList();
			SourcePatchMap = new ArrayList[Source.Width, Source.Height];
			for(int i = 0; i < Source.Width; i++)
				for(int j = 0; j < Source.Height; j++)
					SourcePatchMap[i,j] = new ArrayList();

			PatchSizeMax = patchsizemax;
			if(fixedsize)
				PatchSizeMin = PatchSizeMax;
			else
				PatchSizeMin = 4;
			MaxIterations = maxiterations;
			RandomIterationModulus = randomiterationmodulus;
			QuiltWidth = quiltwidth;
			QuiltHeight = quiltheight;
			PixelErrorMethod = errormethod; // save this string, as we'll use it in status reporting.
			SetPixelErrorDelegate();		// pick which function to use for calculating pixel difference error.

			LatestBitmapOk = false;         // set some flags for UI interaction
			PatchesOk = false;
			StopRequested = false;

			EpitomeSequence = new ArrayList();

			StatString = "Epitome Not Generated";
		}

		// here is where all the work of generating the epitome takes place.
		public void GenerateEpitome()
		{
			NumPatches = 0;
			Patches = new ArrayList();
			int iteration = 0;
			double EpitomeError = 0,  // error for current iteration
				PrevError = 0;     // error for previous iteration
			double score;

			R = new Random();

			int trials = GeneratePatches(Source);  // Generate all the patches from the source image
			RaiseStringEvent("Total Patches: " + NumPatches + " Trials: " + trials + "\n");

			EpitomeBitmap = new Bitmap(Width, Height);
			InitializeBitmap(Source);			// set the initial epitome to noise.
			RaiseBitmapEvent(EpitomeBitmap);	// show the results in the debug window.

			for(int p = 0; p < Patches.Count; p++) // Give all patches an initially random destination.
			{
				((Patch)Patches[p]).DestinationX = R.Next(Width);
				((Patch)Patches[p]).DestinationY = R.Next(Height);
				UpdateEpitomePatchMap(p,MapAdd); // place p into the epitome patch map.
				UpdateSourcePatchMap(p);
				((Patch)Patches[p]).CurScore = PatchScore(p);
				EpitomeError += ((Patch)Patches[p]).CurScore;
			}
			RaiseStringEvent("Initial score = " + EpitomeError + "\n");

			EpitomeError =  PrevError = 0;
			for(iteration = 0; iteration < MaxIterations; iteration++) // for however many iterations were asked for
			{
				EpitomeError = 0.0;
				RaiseStringEvent("Pass " + iteration + ": ");

				// fit the patches
				DateTime StartTime = DateTime.Now;  // note the time at the start of the iteration.
				for(int p = 0; p < Patches.Count; p++) // for each patch...
				{
					double BestScore = ((Patch)Patches[p]).CurScore;  // CurScore is the number to beat
					int BestX = 0, BestY = 0;
					int Trial = 0, MaxTrials = 0, NumPixels = Width * Height;
					int PatchWidth   = 0,  // these variables exist for convenience and self-documentation purposes
						PatchHeight  = 0,
						SearchWidth  = 0,
						SearchHeight = 0;
					int x,y;

					if(p % (Patches.Count/7) == 0) 
						RaiseStringEvent(".");
					if(p % 7 == 0)  // give other people a chance to run too.
						Thread.Sleep(0);

					// For the first few iterations, we just want to try random places for the sake of speed.  But after that, we want to search
					// the vicinity around the so-far-best location.  So, to collapse these two loop constructs into one, we reduce them both to a
					// linear loop over trials, rather than having one loop over some number of random places, and another double-loop over all
					// epitome coordinates.  Then, inside the loop, we determine which case we're in and either choose coordinates randomly or
					// else convert the linear trial number back to epitome coordinates
					if((iteration % RandomIterationModulus) == 0)  // every Nth iteration, do 100 random trials, 
						MaxTrials = Math.Min(100, Width*Height/10);
					else										   // and on the other N-1, search in a box around current pos.
					{
						PatchWidth   = ((Patch)Patches[p]).Width;  // Set up to search a box extending half the patch's size in either
						PatchHeight  = ((Patch)Patches[p]).Height; // direction from the current location.  The + 1 accounts for the
						SearchWidth  = PatchWidth  + 1;            // row and column contining the current location itself.
						SearchHeight = PatchHeight + 1;
						MaxTrials = (SearchWidth) * (SearchHeight);// the SearchWidth x SearchHeight box,
					}
					for(Trial = 0; Trial < MaxTrials; Trial++)
					{
						if((iteration % RandomIterationModulus) == 0) // the try-random-places case
						{
							x = R.Next(Width);
							y = R.Next(Height);
						} 
						else // the search-vicinity case
						{
							if((Trial / SearchWidth) < SearchWidth) // the normal search-the-vicinity part
							{
								x = ((Patch)Patches[p]).DestinationX - PatchWidth/2  + (Trial / SearchWidth);
								y = ((Patch)Patches[p]).DestinationY - PatchHeight/2 + (Trial % SearchHeight);
								if(x < 0) x += Width;  if(x >= Width)  x -= Width; // handle the wraparound
								if(y < 0) y += Height; if(y >= Height) y -= Height; 
							} 
							else // plus a few random trials to try to break out of local minima.
							{
								x = R.Next(Width);
								y = R.Next(Height);
							}
						}
						score = PatchScore(p,x,y); // how well does the patch match here?
						if(score < BestScore) // if that's better than before,
						{
							BestX = x; BestY = y; BestScore = score; // remember the location and new score
						}
					}
					if(BestScore < ((Patch)Patches[p]).CurScore)   // If we found a better match than on the last iteration,
					{
						UpdateEpitomePatchMap(p,MapRemove);			// take the old location out of the patch map
						((Patch)Patches[p]).DestinationX = BestX;	// record the winning location
						((Patch)Patches[p]).DestinationY = BestY;
						UpdateEpitomePatchMap(p,MapAdd);			// add the new location to the patch map
						((Patch)Patches[p]).CurScore = BestScore;	// keep track of the new best score.
					}
					EpitomeError += ((Patch)Patches[p]).CurScore;	// keep a sum of the scores.
				}

				if(iteration == MaxIterations-1 || StopRequested == true) // if this will be the last iteration
					EpitomeBitmap = new Bitmap(Width, Height); // clear it out.

				// re-form the epitome as average of the patches @ their locations
				UpdateEpitome(iteration, (double)MaxIterations/(double)MaxSequenceBitmaps);
				DateTime EndTime = DateTime.Now; // note the time after the iteration is complete

				RaiseBitmapEvent(EpitomeBitmap); // show the results in the debug window.

				// brief debug message
				TimeSpan Span = EndTime - StartTime;
				String TimeString = Span.Hours + "h" + Span.Minutes + "m" + Span.Seconds + "s";
				RaiseStringEvent(" score = " + EpitomeError + " took " + TimeString + "\n");

				if(EpitomeError == PrevError) 
				{	// if this iteration didn't improve anything,
					iteration = MaxIterations;  // then short-circuit the rest of the iterations.
				} 
				else 
					PrevError = EpitomeError;	// remember this score for next time.

				if(StopRequested == true)		// if the user has requested that we stop, then stop.
					iteration = MaxIterations;
			}
			StatString = "Patches[" + PatchSizeMin + "," + PatchSizeMax + "]: " + Patches.Count + 
				" Error[" + PixelErrorMethod+ "]: " + EpitomeError;
//			RaiseStringEvent(GetPatchSizeErrorHistogram());
			RaiseDoneEvent(); // tell the UI we're done.
		}

		// update the epitome bitmap as the average of all patches at their locations.
		private void UpdateEpitome(int iteration, double saveFrequency) 
		{
			float PatchesHere = 0;				// Count of patches that overlap a given position in the epitome.
			float RSum = 0, GSum = 0, BSum = 0; // Sum of R, G, and B for patch pixels that match location Ex, Ey.
			int m, p;							// Indexes into patch map lists and patch list.

			for(int Ex = 0; Ex < Width; Ex++)	// loop over all positions in the epitome
				for(int Ey = 0; Ey < Height; Ey++) 
				{
					PatchesHere = (EpitomePatchMap[Ex, Ey]).Count;
					RSum = GSum = BSum = 0;
					for(m = 0; m < PatchesHere; m++) // check all patches at this location in the patch map
					{
						p = (int)(EpitomePatchMap[Ex,Ey][m]);			// get the patch ID
						int x = Ex-((Patch)Patches[p]).DestinationX;	// x,y = location of Ex,Ey within the patch
						int y = Ey-((Patch)Patches[p]).DestinationY;
						if(((Patch)Patches[p]).DestinationX > Ex)		// check for wraparound
							x += Width;
						if(((Patch)Patches[p]).DestinationY > Ey)
							y += Height;
						RSum += ((Patch)Patches[p]).PixelData.GetPixel(x,y).R;	// add up the color values
						GSum += ((Patch)Patches[p]).PixelData.GetPixel(x,y).G;
						BSum += ((Patch)Patches[p]).PixelData.GetPixel(x,y).B;
					}
					// finally, set the pixel in the epitome to the average of those R,G,B sums
					if(PatchesHere > 0)
						EpitomeBitmap.SetPixel(Ex,Ey,Color.FromArgb((int)(RSum/PatchesHere),    // note: this is a floating-point
							(int)(GSum/PatchesHere),    // divide, with subsequent roundoff
							(int)(BSum/PatchesHere)));
					else
						EpitomeBitmap.SetPixel(Ex,Ey,Color.FromArgb(0,0,0)); // if no patches want this pixel, make it black.
				}
			if((double)iteration >= saveFrequency * (double)(EpitomeSequence.Count))
				EpitomeSequence.Add(new Bitmap(EpitomeBitmap));
		}

		// calculate the fitting "score" for a patch at a given position in the epitome bitmap.  This is the sum of
		// pixel errors for each patch pixel compared to the corresponding epitome pixel, but only for epitome pixels
		// that are actually claimed by any patches; if a pixel is unused, the patch gets to claim it "for free" with
		// zero error.  Individual pixel error methods are implemented by PixelError delegate functions.
		private double PatchScore(int p) 
		{
			return PatchScore(p,((Patch)Patches[p]).DestinationX, ((Patch)Patches[p]).DestinationY);
		}
		private double PatchScore(int p, int X, int Y)
		{
			double Score = 0.0;
			int i, j = 0;
			// the rest of these variables are all just for notational convenience.
			int Ew = Width,
				Eh = Height,
				Pw = ((Patch)Patches[p]).Width,
				Ph = ((Patch)Patches[p]).Height;
			Bitmap b = ((Patch)Patches[p]).PixelData;

			for(i = 0; i < Pw; i++)
				for(j = 0; j < Ph; j++) 
					if(EpitomeBitmap.GetPixel((X+i)%Ew,(Y+j)%Eh).ToArgb() != Color.Black.ToArgb())
						Score += PixelErrorDelegate(b.GetPixel(i,j),EpitomeBitmap.GetPixel((X+i)%Ew,(Y+j)%Eh));
			return Score;
		}

		#region Pixel Error Delegates
		// a delegate encapsulating different run-time methods for pixel error calculation.
		private delegate int PixelError(Color p1, Color p2);
		[NonSerialized] private PixelError PixelErrorDelegate; // the actual delegate to be used; set in the constructor

		// a little data type to hold label/function pairs for pixel error delegates
		private class PixelErrorDelegateStruct 
		{
			private string name;
			public string Name { get { return name; } }
			private PixelError func;
			public PixelError Func { get { return func; } }
			public PixelErrorDelegateStruct(string s, PixelError f) 
			{
				name = s; func = f;
			}
		}
		// A list of pixel error delegates.  Static so we don't have to construct it in the constructor
		private static PixelErrorDelegateStruct[] PixelErrorDelegates = {
			new PixelErrorDelegateStruct("RGB",       new PixelError(PixelErrorRGB)),	// rgb distance
			new PixelErrorDelegateStruct("HSB",       new PixelError(PixelErrorHSB)),	// hsb distance
			new PixelErrorDelegateStruct("dHue^3",    new PixelError(PixelErrorHue)),	// hue-difference cubed
			new PixelErrorDelegateStruct("dSat^3",    new PixelError(PixelErrorSat)),	// saturation-difference^3
			new PixelErrorDelegateStruct("dBright^3", new PixelError(PixelErrorBright)) // brightness-difference^3
		};
		// Return a string array of the delegate labels.  Static so the UI can call it before creating an object
		// of type Epitome.
		public static string[] PixelErrorDelegateLabels
		{
			get { // return a string array of the delegate labels, to display to the user.
				string[] labels = new string[PixelErrorDelegates.Length];
				for(int index = 0; index < PixelErrorDelegates.Length; index++) 
					labels[index] = PixelErrorDelegates[index].Name;
				return labels;
			}
		}
		// Set the actual pixel error delegate variable to whichever delegate function the user wants.  If they
		// ask for something that's not in the list, warn them but give them the first thing instead.
		private void SetPixelErrorDelegate() 
		{
			bool goodmethod = false;
			for(int index = 0; index < PixelErrorDelegates.Length; index++) 
			{
				if(PixelErrorDelegates[index].Name == PixelErrorMethod) 
				{
					PixelErrorDelegate = PixelErrorDelegates[index].Func;
					goodmethod = true;
				}
			}
			if(!goodmethod) // if we didn't find whatever error method the caller wanted,
			{
				// tell the user:
				System.Windows.Forms.MessageBox.Show(
					"Unknown pixel error method: " + PixelErrorMethod + "; using " + PixelErrorDelegates[0].Name, 
					"Unknown Pixel Error Method",
					System.Windows.Forms.MessageBoxButtons.OK);
				// re-call ourselves with something we know will work:
				PixelErrorMethod = PixelErrorDelegates[0].Name;
				SetPixelErrorDelegate();
			}
		}

		// RGB color-distance squared
		private static int PixelErrorRGB(Color p1, Color p2) 
		{
			return (p1.R-p2.R)*(p1.R-p2.R) + (p1.G-p2.G)*(p1.G-p2.G) + (p1.B-p2.B)*(p1.B-p2.B);
		}

		// HSB error = hue_diff^2 + (90*bright_sum)^2 + (90*sat_sum^2)
		private static int PixelErrorHSB(Color p1, Color p2) 
		{
			double Hp, Sp, Bp,	// HSB values for pixels in patch and in E
				He, Se, Be,
				Hdiff, Sdiff, Bdiff,
				d1, d2;
			Hp = p1.GetHue();			He = p2.GetHue();
			Sp = p1.GetSaturation();	Se = p2.GetSaturation();
			Bp = p1.GetBrightness();	Be = p2.GetBrightness();
			if(Hp > He) { double z = Hp; Hp = He; He = z; } // make sure Hp is smaller than He
			d1 = He - Hp;
			d2 = Hp + (360.0 - He);
			Hdiff = Math.Min(d1, d2);	// Hdiff = the shorter path around the circle from Hp to He
			Sdiff = 180*Math.Abs(Sp-Se);			// Math.Abs(Sp-Se);
			Bdiff = 180*Math.Abs(Bp-Be);			// Math.Abs(Bp-Be);
			return (int)(double)((Hdiff*Hdiff)*((Sp+Se)*(Bp+Be)/4) + Bdiff*Bdiff + Sdiff*Sdiff);
		}
		// silly.  cubes of the difference in (hue, saturation, brightness), again scaled by 180.
		private static int PixelErrorHue(Color p1, Color p2) 
		{
			double h1 = p1.GetHue(),
					h2 = p2.GetHue();
			if(h1 > h2) { double z = h1; h1 = h2; h2 = z; }
			double  d1 = h2 - h1,
					d2 = h1 + (360.0 - h2);
			return (int)Math.Pow(Math.Min(d1, d2),3);
		}
		private static int PixelErrorSat(Color p1, Color p2) 
		{
			return (int)Math.Pow(Math.Abs(p1.GetSaturation() - p2.GetSaturation()),3);
		}
		private static int PixelErrorBright(Color p1, Color p2) 
		{
			return (int)Math.Pow(180*Math.Abs(p1.GetBrightness() - p2.GetBrightness()), 3);
		}
		#endregion
		/// <summary>
		/// Generate a random assortment of patches, guaranteed to sample every pixel in the source image
		/// at least once.
		/// </summary>
		/// <param name="Source"></param>
		/// <returns></returns>
		private int GeneratePatches(Bitmap Source) 
		{
			// generate patches:
			// make a mask bitmap, initialized to black
			Bitmap Mask = new Bitmap(Source.Width, Source.Height);
			Color black = Color.FromArgb(0,0,0);
			Color white = Color.FromArgb(255,255,255);

			for(int i = 0; i < Source.Width; i++)
				for(int j = 0; j < Source.Height; j++)
					Mask.SetPixel(i,j,black);

			int hits = 0,                                   // how many pixels in the source have been sampled
				hitlimit = Source.Width * Source.Height,    // how many pixels there are in the source
				blackcount = 0,                             // how many unsampled pixels a given patch would hit
				trials=0;                                   // how many total patch-generation attempts have happened.

			int PatchX, PatchY,           // top left corner of a patch
				SourceW = Source.Width,   // efficiency variables
				SourceH = Source.Height,
				PS;                       // Patch Size
			
			while(hits < hitlimit) 
			{
				trials++;   // count up the patch-generation attempts

				// pick a size for the attempt
				PS = R.Next(PatchSizeMax - PatchSizeMin + 1) + PatchSizeMin;  // random from PatchSizeMin up to PatchSizeMax

				// pick a spot in the source image,
				PatchX = R.Next(SourceW+PS) - PS;  // pick anywhere from (-PS,-PS) to (SourceW, SourceH); the overlap
				PatchY = R.Next(SourceH+PS) - PS;  // makes it easier to get the edges sampled...
				
				// ... but now we have to correct for the overlap, which we don't actually want.  So:
				PatchX = Math.Max(0, PatchX);   // for anything that would overlap the edges of the source
				PatchY = Math.Max(0, PatchY);   // image, adjust the patch location inwards.
				PatchX = Math.Min(PatchX, SourceW - PS);
				PatchY = Math.Min(PatchY, SourceH - PS);  
				
				// see if there are any black pixels in the mask in that patch's area.
				//   while doing so, set them to white and count them.
				blackcount = 0;
				for(int i = 0; i < PS; i++)
					for(int j=0; j < PS; j++)
						if(Mask.GetPixel((PatchX+i) % SourceW, (PatchY+j) % SourceH).ToArgb() == black.ToArgb()) // pixel is black
						{
							blackcount++;
							hits++;
							Mask.SetPixel((PatchX+i) % SourceW, (PatchY+j) % SourceH, white);
						}

				// if there were any black pixels in the patch, note that that spot is a worthwhile patch
				if(blackcount > 0) 
				{
					NumPatches++;
					// create the patch and add it to the Patches collection.
					Patches.Add(new Patch(Source, PatchX, PatchY, PS, PS));
				}
			}
			MakePatchQuilt();  // make and store the patch quilt in case the user wants to see it.

			PatchesOk = true;
			return trials;
		}

		private void MakePatchQuilt() // make a bitmap showing as many patches as will fit into QuiltWidth, QuiltHeight
		{
			PatchQuilt = new Bitmap(QuiltWidth, QuiltHeight);

			int Stride = PatchSizeMax + 1;
			int NumPatches = Math.Min((QuiltWidth/Stride) * (QuiltHeight/Stride),
									   Patches.Count);
			int PatchesPerRow = (QuiltWidth/Stride);
			int p, x, y;

			for(p = 0; p < NumPatches; p++) 
			{
				x = Stride * (p % PatchesPerRow);  // convert the patch number into an x,y offset into b.
				y = Stride * (p / PatchesPerRow); 
				for(int i = 0; i < ((Patch)Patches[p]).Width; i++)  // copy the patch pixels into b.
					for(int j = 0; j < ((Patch)Patches[p]).Height; j++)
						PatchQuilt.SetPixel(x+i, y+j, ((Patch)Patches[p]).PixelData.GetPixel(i,j));
			}
		}
		public Bitmap GetPatches() // return the patch quilt when asked, or a blank bitmap if that's not ok yet.
		{
			if(PatchesOk)
				return PatchQuilt;
			else
				return new Bitmap(QuiltWidth, QuiltHeight);
		}

		// Initialize the epitome to pixels drawn randomly from the source image, then averaged over 3x3 blocks.
		private void InitializeBitmap(Bitmap Source) 
		{
			int i=0, j=0, k=0, l=0;
			int x=0, y=0;
			int r, g, b;
			Bitmap temp = new Bitmap(Width, Height);
			
			for(i=0; i < Width; i++) 
				for(j=0; j < Height; j++) 
				{
					x = R.Next(Source.Width);
					y = R.Next(Source.Height);
					temp.SetPixel(i, j, Source.GetPixel(x,y));
				}
			for(i = 0; i < Width; i++)
				for(j = 0; j < Height; j++) 
				{
					r = g = b = 0;
					for(k = -1; k <= 1; k++)
						for(l = -1; l <= 1; l++)
						{
							x = i+k; y = j+l;
							if(x < 0) x += Width;  if(x >= Width) x -= Width;
							if(y < 0) y += Height; if(y >= Height) y -= Height;
							r += temp.GetPixel(x,y).R;
							g += temp.GetPixel(x,y).G;
							b += temp.GetPixel(x,y).B;
						}
					r /= 9; g /= 9; b /= 9;
					EpitomeBitmap.SetPixel(i,j, Color.FromArgb(r,g,b));
				}
		}

		// return the latest bitmap reported through RaiseBitmapEvent, or else a blank bitmap if none has yet been returned.
		public Bitmap GetLatestBitmap() 
		{
			if(LatestBitmapOk) 
				return LatestBitmap;
			else 
				return new Bitmap(Width, Height);
		}

		// two functions for raising events that will ultimate update the UI.
		private void RaiseStringEvent(string s)				// send a string back to the UI.
		{
			EpitomeStringArgs e = new EpitomeStringArgs(s);
			if(OnEpitomeString != null)
				OnEpitomeString(this, e);
		}
		private void RaiseBitmapEvent(Bitmap b)				// send a bitmap back to the UI
		{
			LatestBitmap = new Bitmap((Image)b, new Size(b.Width, b.Height)); // make a copy so the worker thread can
			EpitomeBitmapArgs e = new EpitomeBitmapArgs(LatestBitmap);        // keep working on whatever bitmap b was.
			LatestBitmapOk = true;
			if(OnEpitomeBitmap != null)
				OnEpitomeBitmap(this, e);
		}
		private void RaiseDoneEvent()						// tell the UI we're done.
		{
			if(OnEpitomeDone != null)
				OnEpitomeDone(this, EventArgs.Empty);
		}
		/// <summary>
		/// Map the epitome image back to an image of the same size as the original input image.  This will be different
		/// than the original image to the extent that the epitome image does not perfectly match all of the patches.
		/// </summary>
		/// <returns></returns>
		public Bitmap MapToImage()
		{
			Bitmap temp = new Bitmap(Source.Width, Source.Height);
			int Px, Py, Ex, Ey;
			int BestPatch = 0,
				BestScore = 0;
			for(int i = 0; i < Source.Width; i++)
				for(int j = 0; j < Source.Height; j++) 
				{
					for(int m = 0; m < SourcePatchMap[i,j].Count; m++) // find the best patch of those that match here.
					{
						int p = (int)(SourcePatchMap[i,j][m]);	// get the patch ID
						if(m == 0 || ((Patch)Patches[p]).CurScore < (int)BestScore) 
						{ 
							BestPatch = p;
							BestScore = (int)((Patch)Patches[p]).CurScore; 
						}
					}
					Px = i - ((Patch)Patches[BestPatch]).OriginX;   // location within the patch with respect to the patch origin.
					Py = j - ((Patch)Patches[BestPatch]).OriginY;
					Ex = (((Patch)Patches[BestPatch]).DestinationX + Px) % EpitomeBitmap.Width;	// location in the epitome
					Ey = (((Patch)Patches[BestPatch]).DestinationY + Py) % EpitomeBitmap.Height;
					// set the mapped pixel to the epitome pixel from the best patch at that location.
					temp.SetPixel(i,j,EpitomeBitmap.GetPixel(Ex, Ey));
				}
			return temp;
		}

		/// <summary>
		/// Return a bitmap showing how many patches match each location in the epitome.  The bitmap is grayscale,
		/// with black = 0 patches, white = highest patch amount.
		/// </summary>
		/// <returns></returns>
		[NonSerialized] private static Color[] DensityColorMap = {
			Color.FromArgb(0,0,0),		
			Color.FromArgb(32,32,32),
			Color.FromArgb(64,64,64),
			Color.FromArgb(96,96,96),
			Color.FromArgb(128,128,128),
			Color.FromArgb(160,160,160),
			Color.FromArgb(192,192,192),
			Color.FromArgb(224,224,224),
			Color.FromArgb(255,255,255)
		};
		[NonSerialized] private static int[] DensityLevels = { // a logarithmic value-to-brightness mapping
			  0,
			  2,
			  4,
			  8,
			 16,
			 32,
			 64,
			128,
			255
		};
		public Bitmap GetPatchDensityMap() 
		{
			Bitmap temp = new Bitmap(Width, Height);
			double MaxPatches, MinPatches;
			int i, j;
			MaxPatches = MinPatches = EpitomePatchMap[0,0].Count;
			for(i = 0; i < Width; i++)
				for(j = 0; j < Height; j++) 
				{
					MaxPatches = Math.Max(EpitomePatchMap[i,j].Count , MaxPatches);
					MinPatches = Math.Min(EpitomePatchMap[i,j].Count , MinPatches);
				}
			if(MaxPatches != MinPatches) // if every spot matches the same number of patches, return the default bitmap
				for(i = 0; i < Width; i++)
					for(j = 0; j < Height; j++) 
					{
						int level = (int)(((double)EpitomePatchMap[i,j].Count-MinPatches)/(MaxPatches-MinPatches) * 255.0);
						int colorindex = 0;
						for(int k = 1; k < DensityLevels.Length; k++)
							if(level <= DensityLevels[k] && 
							   level >  DensityLevels[k-1]) 
								colorindex = k;
						temp.SetPixel(i,j,DensityColorMap[colorindex]);
					}
			RaiseStringEvent("Patch Density Range: " + MinPatches + "," + MaxPatches + "\n");
			return temp;
		}

		/// <summary>
		/// Iterate over epitome coordinates in a box of patch[p]'s size starting from patch[p]'s
		/// destination coordinates.  Add or remove p from the ArrayList in EpitomePatchMap at those coordinates.
		/// This method should be called with action = MapRemove immediately before assigning to a patch's
		/// DestinationX,Y fields, and then called again with action = MapAdd immediately after the assignment.
		/// Note that you need not call with action = MapRemove on the initial assignment to a patch's destination
		/// coordinates.
		/// </summary>
		/// <param name="p"></param>
		private void UpdateEpitomePatchMap(int p, int action) 
		{
			int dx = ((Patch)Patches[p]).DestinationX, // convenience variables
				dy = ((Patch)Patches[p]).DestinationY,
				w  = ((Patch)Patches[p]).Width,
				h  = ((Patch)Patches[p]).Height;
			int ex, ey; // epitome coordinates
			for(int i = 0; i < w; i++) {
				ex = (dx+i) % Width;
				for(int j = 0; j < h; j++) 
				{
					ey = (dy+j) % Height;
					if(action == MapAdd)
						EpitomePatchMap[ex,ey].Add(p);
					else
						EpitomePatchMap[ex,ey].Remove(p);
				}
			}
		}

		/// <summary>
		/// Update a map of patches for locations in the source image, just like we do for the epitome image.
		/// </summary>
		/// <param name="p"></param>
		private void UpdateSourcePatchMap(int p) 
		{
			int ox = ((Patch)Patches[p]).OriginX, // convenience variables
				oy = ((Patch)Patches[p]).OriginY,
				w  = ((Patch)Patches[p]).Width,
				h  = ((Patch)Patches[p]).Height;
			for(int i = 0; i < w; i++)					// loop over the size of the patch,
				for(int j = 0; j < h; j++) 
					SourcePatchMap[ox+i,oy+j].Add(p);	// add p to the relevant lists.
		}

		public void Finish() 
		{
			StopRequested = true;
		}

		public void OnDeserialization(object sender) 
		{
			MakePatchQuilt();
			PatchesOk = true;
			R = new Random();
			StopRequested = false;
		}

//		private string GetPatchSizeErrorHistogram() 
//		{
//			int[] patchcounts = new int[PatchSizeMax+1];
//			double[] scores = new double[PatchSizeMax+1];
//
//			for(int p = 0; p < Patches.Count; p++) 
//			{
//				int size = ((Patch)(Patches[p])).Width;
//				double err = ((Patch)(Patches[p])).CurScore;
//				patchcounts[size]++;
//				scores[size] += err;
//			}
//			string report = "per-pixel error vs patch size:\n";
//			for(int i = PatchSizeMin; i <= PatchSizeMax; i++) 
//				report += "  " + i +":  " + scores[i]/(double)(patchcounts[i]*i*i) + "\n";
//			return report;
//		}
	}
}
